""" Util functions """
from subprocess import PIPE, DEVNULL, run, Popen
from os import environ, path, listdir
from jimpass.config import DEFAULTS
import yaml


def read_config_file(file: str) -> dict:
    """
    Complement context with user configuration
    TODO: schema verification
    """
    with open(file, 'r') as f:
        return yaml.safe_load(f)


def get_config() -> dict:
    """
    Provide configuration based on canonical paths or default
    """
    if 'XDG_CONFIG_HOME' in environ:
        xdg_h = environ['XDG_CONFIG_HOME']
        if path.isdir(f"{xdg_h}/jimpass"):
            for f in listdir(f"{xdg_h}/jimpass"):
                if f.startswith("conf") and f.endswith(".yaml" or ".yml"):
                    return read_config_file("/".join([xdg_h, 'jimpass', f]))
    if 'HOME' in environ:
        for f in listdir(f"{environ['HOME']}"):
            if f.startswith(".jimpass") and f.endswith(".yaml" or ".yml"):
                return read_config_file(f"{environ['HOME']}/{f}")
    return DEFAULTS


def srun(input_cmd: str, stdin: str = None, no_output: bool = False) -> (int, str):
    """
    Wrapper for subprocess_run
    """
    if stdin:
        res = Popen(input_cmd, stdout=DEVNULL if no_output else PIPE, stderr=DEVNULL, stdin=PIPE, shell=True)
        output = None if no_output else res.communicate(bytes(stdin, 'utf-8'))[0].decode()
        return res.returncode, output
    else:
        res = run(input_cmd, stdout=DEVNULL if no_output else PIPE, stderr=DEVNULL, encoding="utf-8", shell=True)
        return res.returncode, res.stdout


def rofi(mode: str = "dmenu", prompt: str = None, options: list = None,
         keybindings: list = None, args: dict = None, stdin: str = None) -> (int, str):
    """
    Run Rofi
    """
    cmd = "rofi "
    if mode:
        cmd += f"-{mode} "
    if prompt:
        cmd += f"-p \"{prompt}\" "
    if options:
        cmd += " ".join([f"-{opt}" for opt in options]) + " "
    if keybindings:
        cmd += " ".join([f"-kb-custom-{kb.exit_code} {kb.mapping}" for kb in keybindings]) + " "
    if args:
        cmd += " ".join([f"-{key} \"{val}\"" for key, val in args.items()]) + " "
    return srun(cmd, stdin) if stdin else srun(cmd)



from jimpass.util import srun, rofi
from jimpass.managers.base import PasswordManager
from jimpass.parser import Parser
import json

item_types = {
    'LOGIN': 1,
    'NOTE': 2,
    'CARD': 3,
    'IDENTITY': 4
}

# How to find fields in the item
login_parser_mapping = {
    'name': 'name',
    'username': 'login.username',
    'id': 'id',
    'password': 'login.password'
}


class Bitwarden(PasswordManager):
    def __init__(self, config: dict):
        PasswordManager.__init__(self, config, 'bitwarden')
        lock_timer = self.config.get("lock_timer")
        if lock_timer:
            self.session_mgr = BitwardenSession(lock_timer, self.config["auto_lock"])
        else:
            self.session_mgr = BitwardenSession(auto_lock=sefl.config["auto_lock"])
        self._parser = Parser(self.pm_config['template_str'], login_parser_mapping)
        self.session = self.session_mgr.get_session()
        self._items = self._fetch_all_items()
        self._full_template_str = "{name}: {username} ({id})"

    def _fetch_all_items(self) -> [dict]:
        """
        Get the items list and filter for logins
        :return: list of all items
        """
        item_str = srun(f"bw list items --session '{self.session}' 2>/dev/null")[1]
        return [self.parser.flat_map_item(item)
                for item in json.loads(item_str, encoding='utf-8')
                if item['type'] == item_types['LOGIN']]

    def sync(self):
        exit_code, _ = srun(f"bw sync --session '{self.session}' &> /dev/null", no_output=True)
        if exit_code == 0:
            self._items = self._fetch_all_items()
        # TODO Warning

    def get_totp(self, item) -> str:
        if 'id' not in item:
            raise Exception("Can't get TOTP from invalid item")
        exit_code, result = srun(f"bw --session '{self.session}' get totp \"{item['id']}\"")
        if exit_code == 0:
            return result
        else:
            return ""


class BitwardenSession(object):
    """
    Manages session key by calling keyctl through subprocess
    """
    def __init__(self, timeout: int = 0, auto_lock: bool = True):
        self.auto_lock = auto_lock
        self.timeout = timeout if auto_lock else -1

    def get_session(self) -> str:
        """ Get the key holding the session hash """
        code, stdout = srun("keyctl request user bw_session")
        if code != 0 or not stdout:
            code, passwd = rofi(
                prompt='Bitwarden Master Password',
                options=[
                    'password'
                ],
                args={
                   'lines': 0
                }
            )
            code, session_key = srun("bw unlock --raw", stdin=passwd)
            if code == 0:
                self.set_session(session_key)
                return session_key
        else:
            return srun(f"keyctl pipe {stdout}")[1]

    def set_session(self, session_key: str):
        """ Set the key holding the session hash """
        if session_key:
            code, key_id = srun("keyctl padd user bw_session @u", stdin=session_key)
            if self.timeout > 0:
                if srun(f"keyctl timeout \"{int(key_id)}\" {self.timeout}")[0] != 0:
                    raise Exception(f"Couldn't set timeout for key_id {int(key_id)}")
            elif self.timeout == 0:
                if srun("keyctl purge user bw_session")[0] != 0:
                    raise Exception(f"Couldn't purge key_id {int(key_id)}")



""" Unit tests live here"""
from jimpass.managers.bitwarden import BitwardenSession
from jimpass.util import srun
import time
import os
import json


class TestKeyManager(object):
    """Testing the Key Manager Class"""

    @property
    def password(self):
        return os.environ["BITWARDEN_PASSWORD"]

    def generate_session(self):
        """
        Generates a new session_key
        :return: the session_key and the key_id for keyctl to find it
        """
        srun("keyctl purge user bw_session &> /dev/null", no_output=True)
        exit_code, session_key = srun("bw unlock --raw", stdin=self.password)
        assert(exit_code == 0)
        assert(len(session_key) > 0)
        exit_code, key_id = srun("keyctl padd user bw_session @u", stdin=session_key)
        assert(exit_code == 0)
        return session_key, key_id

    def test_get_no_autolock(self):
        session, _ = self.generate_session()
        mgr = BitwardenSession(auto_lock=False)
        session_from_mgr: str = mgr.get_session()
        # Since the session was already generated, simply retrieve it
        assert(session_from_mgr == session)

    def test_set_with_autolock(self):
        timeout = 3
        session, _ = self.generate_session()
        mgr = BitwardenSession(auto_lock=True, timeout=timeout)
        srun("keyctl purge user bw_session &> /dev/null", no_output=True)
        mgr.set_session(session)
        time.sleep(timeout+1)
        exit_code, _ = srun("keyctl request user bw_session")
        assert(exit_code != 0)
        # reset keyctl for other functions
        self.generate_session()

    def test_fetch_items(self):
        mgr = BitwardenSession(auto_lock=False)
        session = mgr.get_session()
        resultcode, items_str = srun(f"bw list items --session '{session}' 2> /dev/null")
        items = json.loads(items_str)
        assert(len(items) > 0 and resultcode == 0)



""" Abstract backend """
from abc import ABCMeta, abstractmethod
from deepdiff import DeepDiff
from jimpass.parser import Parser


class PasswordManager(metaclass=ABCMeta):
    """
    Since there is only one backend for now, we will let concrete managers define the flow.
    In the future though, flow should be guided by abstract methods
    """
    def __init__(self, config: dict, name: str):
        """
        :param config:  must include `managers`
        :param name: the password manager's name in the config
        """
        self.config = config
        self.pm_config = config[name]
        self._items = []
        self._parser = None
        self._name = name
        self._full_template_str = "{id}"
        self._current_template_str = None

    @abstractmethod
    def _fetch_all_items(self) -> [dict]:
        pass

    @abstractmethod
    def sync(self):
        """
        Pull the latest update from the remote database
        """
        pass

    @property
    def name(self):
        return self._name

    @property
    def items(self) -> list:
        return self._items

    @property
    def parser(self) -> Parser:
        return self._parser

    @property
    def full_template_str(self) -> str:
        return self._full_template_str

    @staticmethod
    def _is_sub_dict(d1: dict, d2: dict) -> bool:
        """
        :param d1: contained in d2
        :param d2: contains d1
        :return: if d2 contains all fields in d1
        """
        diff = DeepDiff(d1, d2)
        return 'dictionary_item_added' in diff and len(diff) == 1

    def stringify_items(self, items: [dict] = None) -> str:
        """
        Leverage parser to render each database item according to the configured template
        :return: line-separated items to display
        """
        if not items:
            items = self.items
        return '\n'.join([
            self._parser.dumps(item)
            for item in items])

    def search(self, stub: dict) -> [dict]:
        """
        Fetches an item from the database based on a partial dict
        :param stub: dict with partial information to match with items in the db
        :return: a list of matching database items
        """
        return [item
                for item in self.items
                if self._is_sub_dict(stub, item)]

    def get_totp(self, item) -> str:
        """
        Get the item's TOTP 2FA. Subclass implements if supported.
        Default behaviour is empty string
        """
        return ""




from jimpass.parser import Parser

template_str = "account <b>{name}</b> with user {username}"
mapping = {
    'name': 'name',
    'username': 'login.username'
}

parser = Parser(
    template_str,
    mapping
)

dummy_item = {
    "name": "Bob",
    "login": {
        "username": "sponge.bob"
    }
}
dummy_str = "account <b>Bob</b> with user sponge.bob"

dummy_incomplete_item = {
    "name": "Bob"
}
dummy_incomplete_str = "account <b>Bob</b> with user ?"  # intentionally missing


def test_dump():
    assert(parser.dumps(parser.flat_map_item(dummy_item)) == dummy_str)


def test_load():
    loaded_item = parser.loads(dummy_str)
    for field in loaded_item.keys():
        assert(field in mapping)
    loaded_incomplete_item = parser.loads(dummy_incomplete_str)
    assert(loaded_incomplete_item['name'] == dummy_incomplete_item['name'])



from parse import parse


class Parser(object):
    """
    Converts items in a password manager's database from/to the user's template string
    based on a model
    """
    def __init__(self, template_str: str, mapping: dict):
        """
        :param template_str: retrieved from configuration
        eg. "account <b>{name}</b> with user {username}"
        :param mapping: maps parameters in template_str to an field in the database item
        eg. {'name': 'name', 'username': 'login.username'}
        """
        self._template_str = template_str
        self.mapping = mapping

    @property
    def template_str(self) -> str:
        return self._template_str

    @template_str.setter
    def template_str(self, template_str):
        self._template_str = template_str

    def fetch_param_from_dict(self, item: dict, param_name: str) -> str:
        if param_name not in self.mapping.keys():
            raise Exception("Requested parameter is not mapped to a field in item dictionary"
                            f"Available mappings are:\n {','.join(self.mapping.keys())}")
        param_path = self.mapping[param_name].split('.')
        sub = item
        for p in param_path:
            if p not in sub:
                return "?"
            sub = sub[p]
        return str(sub) if sub else "?"

    def dumps(self, item: dict) -> str:
        return self.template_str.format(**item)

    def flat_map_item(self, original_item: dict) -> dict:
        """
        Make a flat map out of an item
        :param original_item: fetched from the database
        :return: its flat map according to `login_parser_mapping`
        """
        new_item = {}
        for k in self.mapping.keys():
            new_item[k] = self.fetch_param_from_dict(original_item, k)
        return new_item

    def loads(self, item_str: str) -> dict:
        params = parse(self.template_str, item_str).named
        for key in params.keys():
            if key not in self.mapping:
                raise Exception("Unkown variable in template string")
        res = {}
        for param_name, val in params.items():
            if val not in '?':
                res[param_name] = val
        return res

    def str_matches_mapping(self, item_str: str) -> bool:
        """
        Checks if string matches the password manager's template str
        :param item_str:
        :return:
        """
        parse_result = parse(self.template_str, item_str)
        if not parse_result:
            return False
        params = parse_result.named
        for key in params.keys():
            if key not in self.mapping:
                return False
        return True




"""
Controller.
Interface with both Password Managers and Rofi
"""
import time
from typing import NamedTuple
from jimpass.managers.base import PasswordManager
from jimpass.util import rofi, srun

COPY_COMMANDS = {
    'xclip': {
        'get': 'xclip -selection clipboard -o',
        'set': 'xclip -selection clipboard -r',
        'clear': 'echo -n "" | xclip -selection clipboard -r'
    },
    'xsel': {
        'get': 'xsel --clipboard',
        'set': 'xsel --clipboard --input',
        'clear': 'xsel --clipboard --clear'
    },
    'wayland': {
        'get': 'wl-paste',
        'set': 'wl-copy',
        'clear': 'wl-copy --clear'
    }
}


class Controller:
    """
    Instantiates password managers and handles user interaction
    """
    def __init__(self, config: dict, managers: dict):
        self.managers = {}
        self.config = config
        self.context = 'base'
        self.keybindings: [KeyBinding] = [
            KeyBinding(exit_code=0, callback='copy_pass')
        ]
        self.copy_command = COPY_COMMANDS[self.config['copy_command']]
        for callback, mapping in config['keybindings'].items():
            non_overlapping_exit_code = len(self.keybindings)
            self.keybindings.append(
                KeyBinding(exit_code=non_overlapping_exit_code,
                           mapping=mapping,
                           callback=callback)
            )
        for name in self.config['managers']:
            self.managers[name] = managers[name](self.config)

    def _get_keybinding(self, exit_code: int) -> int:
        for k in self.keybindings:
            if k.exit_code == exit_code:
                return self.keybindings.index(k)
        # Default to 0 in case of doubt
        return 0

    def _output_switcher(self,
                         exit_code: int,
                         item_list: [dict],
                         mgr: PasswordManager):
        """
        Choose the method to execute based on its name
        When not 0, rofi can be configured to return exit codes from 10 - 18
        corresponding to kb-custom-1 - 9
        :param exit_code: the exit code returned by rofi
        :param item_list: database items from a single password manager
        :param mgr: the Password Manager
        :return:
        """
        if exit_code > 0:
            exit_code -= 9

        key_bind = self.keybindings[self._get_keybinding(exit_code)]
        callback = getattr(self,
                           f"_{key_bind.callback}",
                           lambda: 'Method does not exist!')
        if len(item_list) == 1:
            return callback(item_list[0], mgr)
        return callback(
            self._deduplicate(item_list, mgr)[1] if item_list else None,
            mgr)

    def _type_all(self, item, mgr):
        self._type_user(item, mgr)
        srun("xdotool key Tab")
        self._type_pass(item, mgr)

    @staticmethod
    def _type_user(item, mgr):
        time.sleep(0.2)
        username = item["username"]
        srun(f"xdotool type '{username}'")

    def _type_pass(self, item, mgr):
        time.sleep(0.2)
        password = item["password"]
        srun(f"xdotool type '{password}'")
        if self.config['danger_mode']:
            srun("xdotool key Return")

    def _copy_pass(self, item, mgr):
        password = item["password"]
        srun(f"echo -n '{password}' | {self.copy_command['set']}", no_output=True)
        time.sleep(self.config['clipboard_timeout'])
        if srun(self.copy_command['get'])[1] == password:
            srun(self.copy_command['clear'], no_output=True)

    def _copy_totp(self, item, mgr):
        totp = mgr.get_totp(item)
        if len(totp) > 0:
            srun(f"echo -n '{totp}' | {self.copy_command['set']}", no_output=True)

    def _sync(self, item, mgr):
        for name, m in self.managers.items():
            m.sync()
        self.show_items()

    def _generate_instructions(self) -> str:
        return " | ".join(
            [f"{kb.callback}: {kb.mapping}" for kb in self.keybindings]
        )

    def show_items(self):
        """
        Show items using rofi and return its reponse code and the parsed item
        :return:
        """
        rofi_input = '\n'.join([mgr.stringify_items()
                                for name, mgr in self.managers.items()])
        exit_code, response = rofi(prompt="Name",
                                   keybindings=self.keybindings,
                                   options=['i', 'no-custom'],
                                   args={'mesg': f"{self._generate_instructions()}"},
                                   stdin=rofi_input)
        # if response and exit_code in self.exit_code_to_output.keys():
        if response:
            result_list = [(mgr, mgr.search(mgr.parser.loads(response)))
                           for name, mgr in self.managers.items()
                           if mgr.parser.str_matches_mapping(response)]
            if len(result_list) != 1:
                raise Exception("The item found in more than one Manager")
            found_items = result_list[0][1]
            mgr = result_list[0][0]
            self._output_switcher(exit_code,
                                  found_items,
                                  mgr)

    def _deduplicate(self, items: [dict], mgr: PasswordManager) -> (int, dict):
        mgr.parser.template_str = mgr.full_template_str
        rofi_input = mgr.stringify_items(items)
        exit_code, response = rofi(prompt="Deduplicate",
                                   keybindings=self.keybindings,
                                   options=['i', 'no-custom'],
                                   args={'mesg': self._generate_instructions()+"\n Entries duplicated!"},
                                   stdin=rofi_input)
        if response:
            item = mgr.search(mgr.parser.loads(response))
            return exit_code, item
        return exit_code, None


class KeyBinding(NamedTuple):
    exit_code: int
    callback: str
    mapping: str = "Return"



from jimpass.util import srun
from jimpass.managers.base import PasswordManager
from jimpass.parser import Parser

parser_mapping = {
    'id': 'id',
    'name': 'name',
    'fullname': 'fullname',
    'username': 'username',
    'password': 'password',
    'last_modified': 'last_modified_gmt',
    'last_touch': 'last_touch',
    'group': 'group',
    'url': 'url',
    'note': 'note'
}


class Lastpass(PasswordManager):
    def __init__(self, config: dict):
        PasswordManager.__init__(self, config, 'lastpass')
        self._parser = Parser(self.pm_config['template_str'], parser_mapping)
        self._do_log_in()
        self._items = self._fetch_all_items()
        self._full_template_str = "{name} ({fullname}): {username}, {url}, {group}, ({id})"

    def _do_log_in(self):
        exit_code, _ = srun("lpass status -q", no_output=True)
        if exit_code == 1:
            srun(f"lpass login --trust {self.pm_config['username']}")
        elif exit_code == 0:
            return
        else:
            raise Exception("Couldn't use LastPass, verify it's installed")

    def _fetch_all_items(self) -> [dict]:
        """
        Lastpass supports exporting items into CSV
        :return: flat list of items converted into dict
        """
        exit_code, items_csv = srun("lpass export --fields=id,name,fullname,username,"
                                    "password,last_modified_gmt,last_touch,group,url,note")
        if exit_code != 0:
            raise Exception("Something went wrong retrieving Lastpass items")
        keys = items_csv.split('\n')[0]
        lines = items_csv.split('\n')[1:]
        items = [
            {k: self.parser.flat_map_item(v) for k, v in zip(keys.split(','), line.split(','))}
            for line in lines
        ]
        return items

    def sync(self):
        exit_code, _ = srun("lpass sync", no_output=True)
        if exit_code == 0:
            self._items = self._fetch_all_items()
        # TODO Warning




""" Setup.py """
from setuptools import setup

setup(
    setup_requires=[
        "pytest-runner",
        "pbr"
    ],
    pbr=True,
    tests_require=[
        "pytest"
    ],
    py_modules=['jimpass'],
)



""" Python Bitwarden Rofi """
import yaml
import click
from jimpass.controller import Controller
from jimpass.managers.bitwarden import Bitwarden
from jimpass.managers.lastpass import Lastpass
from jimpass.util import get_config
from jimpass.config import DEFAULTS

MANAGERS = {
    'bitwarden': Bitwarden,
    'lastpass': Lastpass
}


def populate_option(param_, config_key: str, config):
    """
    Priority order:
    1. user input directly from CLI
    2. configuration file
    3. DEFAULTS fallback
    :param param_:
    :param config_key:
    :param config:
    :return:
    """
    if param_:
        config[config_key] = param_
    elif config_key not in config:
        config[config_key] = DEFAULTS[config_key]


@click.command()
@click.option('-c', '--config', type=click.File('r'),
              help='Path to the config file. Defaults to $HOME/.jimpass.yaml or $XDG_CONFIG_HOME/jimpass/config.yaml')
@click.option('--auto-lock', 'lock', flag_value="lock",
              help='Incompatible with --no-auto-lock')
@click.option('--no-auto-lock', 'lock', flag_value="nolock",
              help='Incompatible with --auto-lock')
@click.option('--lock-timer', type=click.IntRange(min=0),
              help='Lock the database after RANGE seconds, ignored if --no-auto-lock')
@click.option('--clipboard-timeout', type=click.IntRange(min=0),
              help='RANGE seconds before clearing up the clipboard after copying')
@click.option('--copy-command', type=click.Choice(['xclip', 'xset', 'wayland']),
              help='Utility to use when copying')
@click.option('--danger-mode', is_flag=True, default=False,
              help='Automatically press Return after auto-type')
def cli(config, lock, lock_timer, clipboard_timeout, copy_command, danger_mode):
    """
    CLI Entrypoint
    """
    if not config:
        config = get_config()
    else:
        config = yaml.load(config)
    # auto-lock
    if lock:
        if lock == "lock":
            config['auto_lock'] = True
        else:
            config['auto_lock'] = False
    elif 'auto_lock' not in config:
        config['auto_lock'] = DEFAULTS['auto_lock']
    # lock-timer
    if 'auto_lock' in config and config['auto_lock']:
        if lock_timer:
            config['lock_timer'] = lock_timer
        elif 'lock_timer' not in config:
            config['lock_timer'] = DEFAULTS['lock_timer']
    # clipboard
    populate_option(clipboard_timeout, 'clipboard_timeout', config)
    # copy-command
    populate_option(copy_command, 'copy_command', config)
    # danger-mode
    populate_option(danger_mode, 'danger_mode', config)
    # Run
    if 'keybindings' not in config:
        config['keybindings'] = DEFAULTS['keybindings']
    Controller(config, MANAGERS).show_items()



DEFAULTS = {
    'managers': [],
    'lock_timer': -1,
    'auto_lock': False,
    'clipboard_timeout': 5,
    'copy_command': 'xclip',
    'danger_mode': False,
    'keybindings': {
        'type_all': 'alt+1',
        'type_user': 'alt+2',
        'type_pass': 'alt+3',
        'copy_totp': 'alt+t',
        'sync': 'alt+r'
    },
    'message': 'Default config loaded'
}
